"""This module contains the version information for BQSKit."""
from __future__ import annotations
__version_info__ = ('1', '1', '2')
__version__ = '.'.join(__version_info__[:3]) + ''.join(__version_info__[3:])
